package tf.bug.loaf

import eu.timepit.refined._
import eu.timepit.refined.api._
import eu.timepit.refined.auto._
import eu.timepit.refined.collection._
import eu.timepit.refined.generic._
import eu.timepit.refined.numeric._
import singleton.ops._

trait Minorable[A <: Matrix[_, _]] {

  type M <: Int
  type N <: Int

  def minor(m: A, i: Int Refined Interval.OpenClosed[0, M], j: Int Refined Interval.OpenClosed[0, N]): Matrix[M - 1, N - 1]

  def zipElementsWithMinors(m: A): LazyList[(Double, Matrix[M - 1, N - 1])]

}

object Minorable {

  def apply[M <: Matrix[_, _] : Minorable]: Minorable[M] = implicitly[Minorable[M]]

  type Aux[Mx <: Int, Nx <: Int, A <: Matrix[Mx, Nx]] = Minorable[A] {
    type M = Mx
    type N = Nx
  }

  trait Implicits {

    implicit def matrixIsMinorable[Mx <: Int : ValueOf, Nx <: Int : ValueOf](implicit bound1: Require[Mx > 1], bound2: Require[Nx > 1]): Minorable.Aux[Mx, Nx, Matrix[Mx, Nx]] = new Minorable[Matrix[Mx, Nx]] {

      override type M = Mx
      override type N = Nx

      override def minor(m: Matrix[M, N], i: Int Refined Interval.ClosedOpen[0, M], j: Int Refined Interval.ClosedOpen[0, N]): Matrix[M - 1, N - 1] = {
        val unboxed = m.columsOfVectors.value.map(_.value)
        val dropRow = unboxed.map(_.zipWithIndex.filter { case (_, rowIndex) => rowIndex != i.value }.map { case (e, _) => e })
        val dropVec = dropRow.zipWithIndex.filter { case (_, colIndex) => colIndex != j.value }.map { case (e, _) => e }
        val reBoxedVecs: List[SList[Double, M - 1]] = dropVec.map(v => refineV[Size[Equal[M - 1]]](v).toOption.get)
        val reBoxed: SList[SList[Double, M - 1], N - 1] = refineV[Size[Equal[N - 1]]](reBoxedVecs).toOption.get
        Matrix(reBoxed)
      }

      override def zipElementsWithMinors(m: Matrix[M, N]): LazyList[(Double, Matrix[M - 1, N - 1])] = {
        LazyList.from(0).take(valueOf[N]).flatMap { i =>
          val ri = refineV[Interval.ClosedOpen[0, M]](i).toOption.get
          LazyList.from(0).take(valueOf[M]).map { j =>
            val rj = refineV[Interval.ClosedOpen[0, N]](j).toOption.get
            (m(ri, rj), minor(m, ri, rj))
          }
        }
      }

    }

  }

}
