package tf.bug.loaf

import singleton.ops._

sealed trait Square[M <: Matrix[_, _]] {

  type Dim <: Int

  def determinant(m: M): Double

}

object Square {

  def apply[M <: Matrix[_, _] : Square]: Square[M] = implicitly[Square[M]]

  trait Implicits {

    implicit def matrixIsSquare[M <: Int : ValueOf, N <: Int : ValueOf](implicit eq: Require[M == N]): Square[Matrix[M, N]] = {
      val d: Int = valueOf[M]
      new Square[Matrix[M, N]] {

        override type Dim = M

        override def determinant(m: Matrix[M, N]): Double = if(d == 1) {
          m.columsOfVectors.value(0).value(0)
        } else {
          val minorable = Minorable[Matrix[M, N]]
          val minorIsSquare = matrixIsSquare[M - 1, N - 1]
          val topElementsLowerMinors = minorable.zipElementsWithMinors(m).take(d)
          topElementsLowerMinors.grouped(2).map(_.toList).map {
            case List((pcoeff: Double, pmatr), (ncoeff: Double, nmatr)) =>
              pcoeff * minorIsSquare.determinant(pmatr) - ncoeff * minorIsSquare.determinant(nmatr)
            case List((pcoeff: Double, pmatr)) =>
              pcoeff * minorIsSquare.determinant(pmatr)
          }.sum
        }

      }
    }

  }

}
